@@ -3,7 +3,7 @@ import ast
 import inspect
 import itertools
 import logging
-from typing import TYPE_CHECKING
+from typing import TYPE_CHECKING, cast
 
 from IPython import get_ipython
 try:
@@ -13,6 +13,7 @@ except ImportError:
 
 from nbsafety.analysis.attr_symbols import AttrSubSymbolChain, CallPoint
 from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
+from nbsafety.data_model.timestamp import Timestamp
 from nbsafety.singletons import nbs
 
 if TYPE_CHECKING:
@@ -87,10 +88,10 @@ class Scope:
         return ret
 
     @staticmethod
-    def _get_name_to_obj_mapping(obj, dc) -> Dict[SupportedIndexType, Any]:
+    def _get_name_to_obj_mapping(obj: Any, dsym: DataSymbol) -> Dict[SupportedIndexType, Any]:
         if obj is None:
             return get_ipython().ns_table['user_global']
-        elif dc is not None and dc.is_subscript:
+        elif dsym is not None and dsym.is_subscript:
             return obj
         else:
             try:
@@ -108,35 +109,62 @@ class Scope:
                 return dict(inspect.getmembers(obj))
         return name_to_obj
 
-    def get_most_specific_data_symbol_for_attrsub_chain(self, chain: AttrSubSymbolChain):
+    def gen_data_symbols_for_attrsub_chain(self, chain: AttrSubSymbolChain):
         """
-        Get most specific DataSymbol for the whole chain (stops at first point it cannot find nested, e.g. a CallPoint).
+        Generates progressive symbols appearing in an AttrSub chain until
+        this can no longer be done semi-statically (e.g. because one of the
+        chain members is a CallPoint).
         """
         cur_scope = self
-        dsym, next_dsym, success = None, None, False
         obj = None
-        for name in chain.symbols:
+        for i, name in enumerate(chain.symbols):
             if isinstance(name, CallPoint):
                 next_dsym = cur_scope.lookup_data_symbol_by_name(name.symbol)
+                if next_dsym is not None:
+                    yield next_dsym, True, False
                 break
             next_dsym = cur_scope.lookup_data_symbol_by_name(name)
-            if dsym is not None and next_dsym is None:
-                # HUGE HACK: prevents us from checking namespace symbols unless entire namespace is stale
-                # TODO: get rid of this check once namespace symbols created for dictionary literals
-                if dsym.is_stale and dsym.timestamp >= dsym.required_timestamp:
-                    dsym = None
+            if next_dsym is None:
                 break
-            dsym, next_dsym = next_dsym, None
+            else:
+                yield next_dsym, False, i == len(chain.symbols) - 1
             try:
-                obj = Scope._get_name_to_obj_mapping(obj, dsym)[name]
+                obj = self._get_name_to_obj_mapping(obj, next_dsym)[name]
             except (KeyError, IndexError, Exception):
                 break
             cur_scope = nbs().namespaces.get(id(obj), None)
             if cur_scope is None:
                 break
-        else:
-            success = True
-        return dsym, next_dsym, success
+
+    def get_most_specific_data_symbol_for_attrsub_chain(self, chain: AttrSubSymbolChain):
+        """
+        Get most specific DataSymbol for the whole chain (stops at first point it cannot find nested, e.g. a CallPoint).
+        """
+        dsym, is_called, success = None, False, False
+        for dsym, is_called, success in self.gen_data_symbols_for_attrsub_chain(chain):
+            pass
+        return dsym, is_called, success
+        # cur_scope = self
+        # dsym, next_dsym, success = None, None, False
+        # obj = None
+        # for name in chain.symbols:
+        #     if isinstance(name, CallPoint):
+        #         next_dsym = cur_scope.lookup_data_symbol_by_name(name.symbol)
+        #         break
+        #     next_dsym = cur_scope.lookup_data_symbol_by_name(name)
+        #     if dsym is not None and next_dsym is None:
+        #         break
+        #     dsym, next_dsym = next_dsym, None
+        #     try:
+        #         obj = Scope._get_name_to_obj_mapping(obj, dsym)[name]
+        #     except (KeyError, IndexError, Exception):
+        #         break
+        #     cur_scope = nbs().namespaces.get(id(obj), None)
+        #     if cur_scope is None:
+        #         break
+        # else:
+        #     success = True
+        # return dsym, next_dsym, success
 
     @staticmethod
     def _resolve_symbol_type(
@@ -201,7 +229,11 @@ class Scope:
             implicit=implicit,
         )
         dsym.update_deps(
-            deps, prev_obj=prev_obj, overwrite=overwrite, propagate=propagate, refresh=not implicit
+            deps,
+            prev_obj=prev_obj,
+            overwrite=overwrite,
+            propagate=propagate,
+            refresh=not implicit,
         )
         return dsym
 
@@ -262,9 +294,10 @@ class Scope:
         return dsym, prev_dsym, prev_obj
 
     def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
+        assert not is_subscript
         dsym = self._data_symbol_by_name.pop(name, None)
         if dsym is not None:
-            dsym.update_deps(set())
+            dsym.update_deps(set(), deleted=True)
 
     @property
     def is_global(self):
@@ -311,6 +344,8 @@ class Scope:
 class NamespaceScope(Scope):
     ANONYMOUS = '<anonymous_namespace>'
 
+    PENDING_CLASS_PLACEHOLDER = object()
+
     # TODO: support (multiple) inheritance by allowing
     #  NamespaceScopes from classes to clone their parent class's NamespaceScopes
     def __init__(self, obj: Any, *args, **kwargs):
@@ -325,10 +360,11 @@ class NamespaceScope(Scope):
                 raise ValueError(msg)
             else:
                 logger.warning(msg)
-        nbs().namespaces[id(obj)] = self
+        if obj is not self.PENDING_CLASS_PLACEHOLDER:
+            nbs().namespaces[id(obj)] = self
         self._tombstone = False
         # this timestamp needs to be bumped in DataSymbol refresh()
-        self.max_descendent_timestamp = -1
+        self.max_descendent_timestamp: Timestamp = Timestamp.uninitialized()
         self._subscript_data_symbol_by_name: Dict[SupportedIndexType, DataSymbol] = {}
         self.namespace_stale_symbols: Set[DataSymbol] = set()
 
@@ -415,7 +451,6 @@ class NamespaceScope(Scope):
             return name
 
     def lookup_data_symbol_by_name_this_indentation(self, name, is_subscript=None, skip_cloned_lookup=False):
-        # TODO: specify in arguments whether `name` refers to a subscript
         if is_subscript is None:
             ret = self._data_symbol_by_name.get(name, None)
             if ret is None:
@@ -430,6 +465,27 @@ class NamespaceScope(Scope):
                 ret = self.cloned_from.lookup_data_symbol_by_name_this_indentation(name, is_subscript=is_subscript)
         return ret
 
+    def _remap_sym(self, from_idx: int, to_idx: int, prev_obj: Optional[Any]) -> None:
+        subsym = self._subscript_data_symbol_by_name.pop(from_idx, None)
+        if subsym is None:
+            return
+        subsym.name = to_idx
+        subsym.invalidate_cached()  # ensure we bypass equality check and bump timestamp
+        subsym.update_deps(
+            set(), prev_obj, overwrite=False, propagate=True, refresh=True,
+        )
+        self._subscript_data_symbol_by_name[to_idx] = subsym
+
+    def shuffle_symbols_upward_from(self, pos: int) -> None:
+        for idx in range(len(self.obj) - 1, pos, -1):
+            prev_obj = self.obj[idx + 1] if idx < len(self.obj) - 1 else None
+            self._remap_sym(idx - 1, idx, prev_obj)
+
+    def _shuffle_symbols_downward_to(self, pos: int) -> None:
+        for idx in range(pos + 1, len(self.obj)):
+            prev_obj = self.obj[idx - 2] if idx > pos + 1 else None
+            self._remap_sym(idx, idx - 1, prev_obj)
+
     def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
         if is_subscript:
             dsym = self._subscript_data_symbol_by_name.pop(name, None)
@@ -438,6 +494,8 @@ class NamespaceScope(Scope):
                 dsym = self._subscript_data_symbol_by_name.pop(name, None)
             if dsym is not None:
                 dsym.update_deps(set(), deleted=True)
+            if isinstance(self.obj, list) and isinstance(name, int):
+                self._shuffle_symbols_downward_to(name)
         else:
             super().delete_data_symbol_for_name(name)
 
@@ -463,8 +521,8 @@ class NamespaceScope(Scope):
             self._data_symbol_by_name[name] = val
         val.containing_scope = self
 
-    def refresh(self):
-        self.max_descendent_timestamp = nbs().cell_counter()
+    def refresh(self) -> None:
+        self.max_descendent_timestamp = Timestamp.current()
 
     def get_earliest_ancestor_containing(self, obj_id: int, is_subscript: bool) -> Optional[NamespaceScope]:
         # TODO: test this properly
