@@ -6,7 +6,7 @@ from typing import TYPE_CHECKING
 from IPython import get_ipython
 
 if TYPE_CHECKING:
-    from typing import List, Optional, Union, Type
+    from typing import Callable, List, Optional
 
 
 def _ipython():
@@ -24,17 +24,19 @@ class _IpythonState:
         self.cell_counter = None
 
     @contextmanager
-    def ast_transformer_context(
-            self, transformers: Union[List[Union[ast.NodeTransformer, Type]], ast.NodeTransformer, Type]
-    ):
-        if not isinstance(transformers, list):
-            transformers = [transformers]
-        transformers = [t if isinstance(t, ast.NodeTransformer) else t() for t in transformers]
+    def ast_transformer_context(self, transformers: List[ast.NodeTransformer]):
         old = _ipython().ast_transformers
         _ipython().ast_transformers = old + transformers
         yield
         _ipython().ast_transformers = old
 
+    @contextmanager
+    def input_transformer_context(self, transformers: List[Callable[[List[str]], List[str]]]):
+        old = _ipython().input_transformers_post
+        _ipython().input_transformers_post = old + transformers
+        yield
+        _ipython().input_transformers_post = old
+
 
 _IPY = _IpythonState()
 
@@ -57,3 +59,7 @@ def run_cell(cell, **kwargs):
     return _ipython().run_cell(
         cell, store_history=kwargs.pop('store_history', True), silent=kwargs.pop('silent', False)
     )
+
+
+class CellNotRunYetError(ValueError):
+    pass
