@@ -1,5 +1,6 @@
 # -*- coding: future_annotations -*-
 import ast
+import astunparse
 import asyncio
 from collections import defaultdict
 from contextlib import contextmanager
@@ -18,8 +19,8 @@ from IPython.core.magic import register_cell_magic, register_line_magic
 
 from nbsafety.analysis.live_refs import (
     compute_live_dead_symbol_refs,
-    compute_call_chain_live_symbols_and_cells,
     get_symbols_for_references,
+    get_live_symbols_and_cells_for_references,
 )
 from nbsafety.ipython_utils import (
     CellNotRunYetError,
@@ -31,16 +32,19 @@ from nbsafety.ipython_utils import (
 from nbsafety import line_magics
 from nbsafety.data_model.data_symbol import DataSymbol
 from nbsafety.data_model.scope import Scope, NamespaceScope
+from nbsafety.data_model.timestamp import Timestamp
 from nbsafety.run_mode import SafetyRunMode
 from nbsafety import singletons
 from nbsafety.tracing.safety_ast_rewriter import SafetyAstRewriter
 from nbsafety.tracing.trace_manager import TraceManager
 
 if TYPE_CHECKING:
-    from typing import Any, Dict, Iterable, List, Set, Optional, Tuple, Union
+    from typing import Any, Dict, Iterable, List, Set, Optional, Tuple, TypeVar, Union
     from types import FrameType
     from nbsafety.types import CellId, SupportedIndexType
 
+    TimestampOrCounter = TypeVar('TimestampOrCounter', Timestamp, int)
+
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.WARNING)
 
@@ -57,7 +61,6 @@ class NotebookSafetySettings(NamedTuple):
     mark_stale_symbol_usages_unsafe: bool
     mark_typecheck_failures_unsafe: bool
     mark_phantom_cell_usages_unsafe: bool
-    naive_refresher_computation: bool
     mode: SafetyRunMode
 
 
@@ -66,6 +69,7 @@ class MutableNotebookSafetySettings:
     trace_messages_enabled: bool
     highlights_enabled: bool
     static_slicing_enabled: bool
+    dynamic_slicing_enabled: bool
 
 
 class CheckerResult(NamedTuple):
@@ -90,13 +94,13 @@ class NotebookSafety(singletons.NotebookSafety):
             mark_stale_symbol_usages_unsafe=kwargs.pop('mark_stale_symbol_usages_unsafe', True),
             mark_typecheck_failures_unsafe=kwargs.pop('mark_typecheck_failures_unsafe', False),
             mark_phantom_cell_usages_unsafe=kwargs.pop('mark_phantom_cell_usages_unsafe', False),
-            naive_refresher_computation=False,
             mode=SafetyRunMode.get(),
         )
         self.mut_settings: MutableNotebookSafetySettings = MutableNotebookSafetySettings(
             trace_messages_enabled=kwargs.pop('trace_messages_enabled', False),
             highlights_enabled=kwargs.pop('highlights_enabled', True),
             static_slicing_enabled=kwargs.pop('static_slicing_enabled', True),
+            dynamic_slicing_enabled=kwargs.pop('dynamic_slicing_enabled', True),
         )
         # Note: explicitly adding the types helps PyCharm intellisense
         self.settrace = settrace or sys.settrace
@@ -111,8 +115,10 @@ class NotebookSafety(singletons.NotebookSafety):
         # TODO: we have a lot of fields concerning cells; they should probably get their own
         #  abstraction in the data model via a dedicated class
         self.cell_content_by_counter: Dict[int, str] = {}
+        self.cell_ast_by_counter: Dict[int, ast.Module] = {}
         self.statement_to_func_cell: Dict[int, DataSymbol] = {}
-        self.cell_counter_by_live_symbol: Dict[DataSymbol, Set[CellId]] = defaultdict(set)
+        self.cell_counter_by_live_symbol: Dict[DataSymbol, Set[int]] = defaultdict(set)
+        self.live_symbols_by_cell_counter: Dict[int, Set[DataSymbol]] = defaultdict(set)
         self.cell_counters_needing_typecheck: Set[int] = set()
         self._typecheck_error_cells: Set[CellId] = set()
         self._counter_by_cell_id: Dict[CellId, int] = {}
@@ -130,9 +136,8 @@ class NotebookSafety(singletons.NotebookSafety):
         self._last_refused_code: Optional[str] = None
         self._prev_cell_stale_symbols: Set[DataSymbol] = set()
         self._cell_counter = 1
-        self._recorded_cell_name_to_cell_num = True
         self._cell_name_to_cell_num_mapping: Dict[str, int] = {}
-        self._ast_transformer_raised: Optional[Exception] = None
+        self._exception_raised_during_execution: Optional[Exception] = None
         self._saved_debug_message: Optional[str] = None
         if use_comm:
             get_ipython().kernel.comm_manager.register_target(__package__, self._comm_target)
@@ -160,7 +165,7 @@ class NotebookSafety(singletons.NotebookSafety):
                 return alias
         return None
 
-    def cell_counter(self):
+    def cell_counter(self) -> int:
         if self.settings.store_history:
             return cell_counter()
         else:
@@ -170,14 +175,14 @@ class NotebookSafety(singletons.NotebookSafety):
         # only called in test context
         assert not self.settings.store_history
         for sym in self.all_data_symbols():
-            sym.last_used_cell_num = sym._timestamp = sym._max_inner_timestamp = sym.required_timestamp = 0
+            sym._timestamp = sym._max_inner_timestamp = sym.required_timestamp = Timestamp.uninitialized()
             sym.timestamp_by_used_time.clear()
-            sym.timestamp_by_liveness_time.clear()
+            sym.timestamp_by_liveness_time_by_cell_counter.clear()
         self._cell_counter = 1
 
-    def set_ast_transformer_raised(self, new_val: Optional[Exception] = None) -> Optional[Exception]:
-        ret = self._ast_transformer_raised
-        self._ast_transformer_raised = new_val
+    def set_exception_raised_during_execution(self, new_val: Optional[Exception] = None) -> Optional[Exception]:
+        ret = self._exception_raised_during_execution
+        self._exception_raised_during_execution = new_val
         return ret
 
     def get_position(self, frame: FrameType) -> Tuple[Optional[int], int]:
@@ -261,7 +266,7 @@ class NotebookSafety(singletons.NotebookSafety):
             ):
                 continue
             try:
-                checker_result = self._check_cell_and_resolve_symbols(cell_content)
+                checker_result = self._check_cell_and_resolve_symbols(cell_content, self._counter_by_cell_id[cell_id])
                 stale_symbols, dead_symbols = checker_result.stale, checker_result.dead
                 if len(stale_symbols) > 0:
                     stale_symbols_by_cell_id[cell_id] = stale_symbols
@@ -288,8 +293,8 @@ class NotebookSafety(singletons.NotebookSafety):
                     pass
 
                 if (
-                        self._get_max_timestamp_cell_num_for_symbols(checker_result.live) >
-                        self._counter_by_cell_id.get(cell_id, cast(int, float('inf')))
+                    self._get_max_timestamp_cell_num_for_symbols(checker_result.live) >
+                    self._counter_by_cell_id.get(cell_id, cast(int, float('inf')))
                 ) and cell_id not in stale_cells and cell_id not in self._typecheck_error_cells:
                     fresh_cells.append(cell_id)
             except SyntaxError:
@@ -299,17 +304,9 @@ class NotebookSafety(singletons.NotebookSafety):
         for stale_cell_id in stale_cells:
             stale_syms = stale_symbols_by_cell_id[stale_cell_id]
             refresher_cell_ids: Set[CellId] = set()
-            if self.settings.naive_refresher_computation:
-                refresher_cell_ids = self._naive_compute_refresher_cells(
-                    stale_cell_id,
-                    stale_syms,
-                    content_by_cell_id,
-                    order_index_by_cell_id=order_index_by_cell_id
-                )
-            else:
-                refresher_cell_ids = refresher_cell_ids.union(
-                    *(killing_cell_ids_for_symbol[stale_sym] for stale_sym in stale_syms)
-                )
+            refresher_cell_ids = refresher_cell_ids.union(
+                *(killing_cell_ids_for_symbol[stale_sym] for stale_sym in stale_syms)
+            )
             stale_links[stale_cell_id] = refresher_cell_ids
         stale_link_changes = True
         # transitive closer up until we hit non-stale refresher cells
@@ -348,30 +345,6 @@ class NotebookSafety(singletons.NotebookSafety):
             'refresher_links': refresher_links,
         }
 
-    def _naive_compute_refresher_cells(
-        self,
-        stale_cell_id: CellId,
-        stale_symbols: Set[DataSymbol],
-        cells_by_id: Dict[CellId, str],
-        order_index_by_cell_id: Optional[Dict[CellId, int]] = None
-    ) -> Set[CellId]:
-        refresher_cell_ids: Set[CellId] = set()
-        stale_cell_content = cells_by_id[stale_cell_id]
-        for cell_id, cell_content in cells_by_id.items():
-            if cell_id == stale_cell_id:
-                continue
-            if (order_index_by_cell_id is not None and
-                    order_index_by_cell_id.get(cell_id, -1) >= order_index_by_cell_id.get(stale_cell_id, -1)):
-                continue
-            concated_content = f'{cell_content}\n\n{stale_cell_content}'
-            try:
-                concated_stale_symbols = self._check_cell_and_resolve_symbols(concated_content).stale
-            except SyntaxError:
-                continue
-            if concated_stale_symbols < stale_symbols:
-                refresher_cell_ids.add(cell_id)
-        return refresher_cell_ids
-
     @staticmethod
     def _get_cell_ast(cell):
         lines = []
@@ -384,10 +357,10 @@ class NotebookSafety(singletons.NotebookSafety):
         return ast.parse('\n'.join(lines))
 
     def _get_max_timestamp_cell_num_for_symbols(self, symbols: Set[DataSymbol]) -> int:
-        max_timestamp = -1
+        max_timestamp_cell_num = -1
         for dsym in symbols:
-            max_timestamp = max(max_timestamp, dsym.timestamp)
-        return max_timestamp
+            max_timestamp_cell_num = max(max_timestamp_cell_num, dsym.timestamp.cell_num)
+        return max_timestamp_cell_num
 
     def _build_typecheck_slice(
         self, cell_id: CellId, content_by_cell_id: Dict[CellId, str], checker_result: CheckerResult
@@ -405,21 +378,30 @@ class NotebookSafety(singletons.NotebookSafety):
             content='\n'.join(content_by_cell_id[cell_id] for cell_id in live_cell_ids),
         )
 
-    def _check_cell_and_resolve_symbols(self, cell: Union[ast.Module, str]) -> CheckerResult:
+    def _check_cell_and_resolve_symbols(
+        self, cell: Union[ast.Module, str], cell_ctr: int, update_liveness_time_versions: bool = False
+    ) -> CheckerResult:
+        for dsym in self.live_symbols_by_cell_counter[cell_ctr]:
+            dsym.timestamp_by_liveness_time_by_cell_counter[cell_ctr].clear()
         if isinstance(cell, str):
             cell = self._get_cell_ast(cell)
         live_symbol_refs, dead_symbol_refs = compute_live_dead_symbol_refs(cell, scope=self.global_scope)
-        live_symbols, called_symbols = get_symbols_for_references(live_symbol_refs, self.global_scope)
-        live_symbols_from_calls, live_cells_from_calls = compute_call_chain_live_symbols_and_cells(called_symbols)
-        live_symbols = live_symbols | live_symbols_from_calls
+        if update_liveness_time_versions:
+            get_live_symbols_and_cells_for_references(
+                live_symbol_refs, self.global_scope, cell_ctr, update_liveness_time_versions=True
+            )
+        live_symbols, live_cells_from_calls = get_live_symbols_and_cells_for_references(
+            live_symbol_refs, self.global_scope, cell_ctr, update_liveness_time_versions=False
+        )
         # only mark dead attrsubs as killed if we can traverse the entire chain
         dead_symbols, _ = get_symbols_for_references(
-            dead_symbol_refs, self.global_scope, only_add_successful_resolutions=True
+            dead_symbol_refs, self.global_scope, only_yield_successful_resolutions=True
         )
         stale_symbols = set(dsym for dsym in live_symbols if dsym.is_stale)
+        self.live_symbols_by_cell_counter[cell_ctr] = live_symbols
         return CheckerResult(
             live=live_symbols,
-            live_cells={sym.timestamp for sym in live_symbols},
+            live_cells={sym.timestamp.cell_num for sym in live_symbols},
             live_cells_from_calls=live_cells_from_calls,
             dead=dead_symbols,
             stale=stale_symbols,
@@ -432,7 +414,7 @@ class NotebookSafety(singletons.NotebookSafety):
     def _stale_sym_warning(sym: DataSymbol):
         if not sym.is_stale:
             raise ValueError('Expected node with stale ancestor; got %s' % sym)
-        if sym.timestamp < 1:
+        if not sym.timestamp.is_initialized:
             return
         fresher_symbols = sym.fresher_ancestors
         if len(fresher_symbols) == 0:
@@ -440,10 +422,10 @@ class NotebookSafety(singletons.NotebookSafety):
                 *(ns_stale.fresher_ancestors for ns_stale in sym.namespace_stale_symbols)
             )
         logger.warning(
-            f'`{sym.readable_name}` defined in cell {sym.timestamp} may depend on '
-            f'old version(s) of [{", ".join(f"`{str(dep)}`" for dep in fresher_symbols)}] '
-            f'(latest update in cell {max(dep.timestamp for dep in fresher_symbols)}).'
-            f'\n\n(Run cell again to override and execute anyway.)'
+            f'`{sym.readable_name}` defined in cell {sym.defined_cell_num} '
+            f'(last updated in cell {sym.timestamp.cell_num}) may depend on '
+            f'old version(s) of [{", ".join(f"`{str(dep)}`" for dep in fresher_symbols if not dep.is_anonymous)}] '
+            f'(latest update in cell {max(dep.timestamp.cell_num for dep in fresher_symbols if not dep.is_anonymous)}).'
         )
 
     @staticmethod
@@ -469,7 +451,9 @@ class NotebookSafety(singletons.NotebookSafety):
             cell_ast = self._get_cell_ast(cell)
         except SyntaxError:
             return False
-        checker_result = self._check_cell_and_resolve_symbols(cell_ast)
+        checker_result = self._check_cell_and_resolve_symbols(
+            cell_ast, self.cell_counter(), update_liveness_time_versions=self.mut_settings.static_slicing_enabled
+        )
         stale_symbols, live_symbols, live_cells = checker_result.stale, checker_result.live, checker_result.live_cells
         stale_sym_usage_warning_counter = 0
         phantom_cell_usage_warning_counter = 0
@@ -483,6 +467,8 @@ class NotebookSafety(singletons.NotebookSafety):
                         break
                     self._stale_sym_warning(sym)
                     stale_sym_usage_warning_counter += 1
+                if stale_sym_usage_warning_counter > 0:
+                    logger.warning('\n\n(Run cell again to override and execute anyway.)')
             if self.settings.mark_phantom_cell_usages_unsafe:
                 used_cell_counters_by_cell_id = defaultdict(set)
                 used_cell_counters_by_cell_id[cell_id].add(self.cell_counter())
@@ -516,11 +502,11 @@ class NotebookSafety(singletons.NotebookSafety):
             self._last_refused_code = cell
             return True
 
-        # For each of the live symbols, record their `defined_cell_num`
+        # For each of the live symbols, record their timestamp
         # at the time of liveness, for use with the dynamic slicer.
         for sym in live_symbols:
             self.cell_counter_by_live_symbol[sym].add(self.cell_counter())
-            sym.timestamp_by_liveness_time[self.cell_counter()] = sym.timestamp
+            # sym.timestamp_by_liveness_time[self.cell_counter()] = sym.timestamp
 
         self._last_refused_code = None
         return False
@@ -574,17 +560,17 @@ class NotebookSafety(singletons.NotebookSafety):
                 continue
             for used_time, sym_timestamp_when_used in sym.timestamp_by_used_time.items():
                 if top_level_sym.is_import:
-                    cell_num_to_used_imports[used_time].add(top_level_sym)
+                    cell_num_to_used_imports[used_time.cell_num].add(top_level_sym)
                 else:
                     if sym_timestamp_when_used < used_time:
-                        cell_num_to_dynamic_cell_parents[used_time].add(sym_timestamp_when_used)
-                        cell_num_to_dynamic_inputs[used_time].add(top_level_sym)
-                        cell_num_to_dynamic_cell_children[sym_timestamp_when_used].add(used_time)
-                    cell_num_to_dynamic_outputs[sym_timestamp_when_used].add(top_level_sym)
+                        cell_num_to_dynamic_cell_parents[used_time.cell_num].add(sym_timestamp_when_used.cell_num)
+                        cell_num_to_dynamic_inputs[used_time.cell_num].add(top_level_sym)
+                        cell_num_to_dynamic_cell_children[sym_timestamp_when_used.cell_num].add(used_time.cell_num)
+                    cell_num_to_dynamic_outputs[sym_timestamp_when_used.cell_num].add(top_level_sym)
             if not top_level_sym.is_import:
                 for updated_time in sym.updated_timestamps:
                     # TODO: distinguished between used / unused outputs?
-                    cell_num_to_dynamic_outputs[updated_time].add(top_level_sym)
+                    cell_num_to_dynamic_outputs[updated_time.cell_num].add(top_level_sym)
 
         cell_metadata: Dict[int, Dict[str, Union[List[int], List[str], Dict[str, Dict[str, str]]]]] = {}
         all_relevant_cells = (
@@ -617,10 +603,10 @@ class NotebookSafety(singletons.NotebookSafety):
             }
         return cell_metadata
 
-    def get_cell_dependencies(self, cell_num: int) -> Dict[int, str]:
+    def compute_slice(self, cell_num: int, stmt_level: bool = False) -> Dict[int, str]:
         """
-        Gets a dictionary object of cell dependencies for the last or 
-        currently executed cell.
+        Gets a dictionary object of cell dependencies for the cell with
+        the specified execution counter.
 
         Args:
             - cell_num (int): cell to get dependencies for, defaults to last
@@ -633,30 +619,86 @@ class NotebookSafety(singletons.NotebookSafety):
         if cell_num not in self.cell_content_by_counter.keys():
             raise CellNotRunYetError(f'Cell {cell_num} has not been run yet.')
 
-        dependencies: Set[int] = set()
-        cell_num_to_dynamic_deps: Dict[int, Set[int]] = defaultdict(set)
-        cell_num_to_static_deps: Dict[int, Set[int]] = defaultdict(set)
+        if stmt_level:
+            stmts_by_cell_num = self.compute_slice_stmts(cell_num)
+            stmts_by_cell_num.pop(cell_num, None)
+            ret = {
+                ctr: '\n'.join(astunparse.unparse(stmt).strip() for stmt in stmts)
+                for ctr, stmts in stmts_by_cell_num.items()
+            }
+            ret[cell_num] = self.cell_content_by_counter[cell_num]
+            return ret
+        else:
+            deps: Set[int] = self._compute_slice_impl(cell_num, set(), defaultdict(set), defaultdict(set))
+            return {dep: self.cell_content_by_counter[dep] for dep in deps}
+
+    def compute_slice_stmts(self, cell_num: int) -> Dict[int, List[ast.stmt]]:
+        if cell_num not in self.cell_content_by_counter.keys():
+            raise CellNotRunYetError(f'Cell {cell_num} has not been run yet.')
+
+        deps_stmt: Set[Timestamp] = self._compute_slice_impl(
+            Timestamp(cell_num, -1), set(), defaultdict(set), defaultdict(set)
+        )
+        stmts_by_cell_num = defaultdict(list)
+        seen_stmt_ids = set()
+        for ts in sorted(deps_stmt):
+            if ts.cell_num > cell_num:
+                break
+            stmt = self.cell_ast_by_counter[ts.cell_num].body[ts.stmt_num]
+            stmt_id = id(stmt)
+            if stmt is None or stmt_id in seen_stmt_ids:
+                continue
+            seen_stmt_ids.add(stmt_id)
+            if stmt is not None:
+                stmts_by_cell_num[ts.cell_num].append(stmt)
+        stmts_by_cell_num[cell_num] = list(ast.parse(self.cell_content_by_counter[cell_num]).body)
+        return dict(stmts_by_cell_num)
+
+    def _compute_slice_impl(
+        self,
+        seed_ts: TimestampOrCounter,
+        dependencies: Set[TimestampOrCounter],
+        timestamp_to_dynamic_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
+        timestamp_to_static_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
+    ) -> Set[TimestampOrCounter]:
 
         for sym in self.all_data_symbols():
-            for used_time, sym_timestamp_when_used in sym.timestamp_by_used_time.items():
-                if sym_timestamp_when_used < used_time:
-                    cell_num_to_dynamic_deps[used_time].add(sym_timestamp_when_used)
+            if self.mut_settings.dynamic_slicing_enabled:
+                for used_time, sym_timestamp_when_used in sym.timestamp_by_used_time.items():
+                    if sym_timestamp_when_used < used_time:
+                        if isinstance(seed_ts, Timestamp):
+                            timestamp_to_dynamic_ts_deps[used_time].add(sym_timestamp_when_used)
+                        else:
+                            timestamp_to_dynamic_ts_deps[used_time.cell_num].add(sym_timestamp_when_used.cell_num)
             if self.mut_settings.static_slicing_enabled:
-                for liveness_time, sym_timestamp_when_used in sym.timestamp_by_liveness_time.items():
-                    if sym_timestamp_when_used < liveness_time:
-                        cell_num_to_static_deps[liveness_time].add(sym_timestamp_when_used)
+                for timestamp_by_liveness_time in sym.timestamp_by_liveness_time_by_cell_counter.values():
+                    for liveness_time, sym_timestamp_when_used in list(timestamp_by_liveness_time.items()):
+                        if sym_timestamp_when_used < liveness_time:
+                            if isinstance(seed_ts, Timestamp):
+                                timestamp_to_static_ts_deps[liveness_time].add(sym_timestamp_when_used)
+                            else:
+                                timestamp_to_static_ts_deps[liveness_time.cell_num].add(
+                                    sym_timestamp_when_used.cell_num
+                                )
 
-        self._get_cell_dependencies(
-            cell_num, dependencies, cell_num_to_dynamic_deps, cell_num_to_static_deps
+        # ensure we at least get the static deps
+        self._get_ts_dependencies(
+            seed_ts, dependencies, timestamp_to_dynamic_ts_deps, timestamp_to_static_ts_deps
         )
-        return {num: self.cell_content_by_counter[num] for num in dependencies}
+        if isinstance(seed_ts, Timestamp):
+            for ts in list(timestamp_to_dynamic_ts_deps.keys() | timestamp_to_static_ts_deps.keys()):
+                if ts.cell_num == seed_ts.cell_num:
+                    self._get_ts_dependencies(
+                        ts, dependencies, timestamp_to_dynamic_ts_deps, timestamp_to_static_ts_deps
+                    )
+        return dependencies
 
-    def _get_cell_dependencies(
+    def _get_ts_dependencies(
         self,
-        cell_num: int,
-        dependencies: Set[int],
-        cell_num_to_dynamic_deps: Dict[int, Set[int]],
-        cell_num_to_static_deps: Dict[int, Set[int]],
+        timestamp: TimestampOrCounter,
+        dependencies: Set[TimestampOrCounter],
+        timestamp_to_dynamic_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
+        timestamp_to_static_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
     ) -> None:
         """
         For a given cell, this function recursively populates a set of
@@ -674,27 +716,32 @@ class NotebookSafety(singletons.NotebookSafety):
             None
         """
         # Base case: cell already in dependencies
-        if cell_num in dependencies or cell_num <= 0:
+        if timestamp in dependencies:
+            return
+        if isinstance(timestamp, int) and timestamp < 0:
+            return
+        if isinstance(timestamp, Timestamp) and not timestamp.is_initialized:
             return
 
         # Add current cell to dependencies
-        dependencies.add(cell_num)
+        dependencies.add(timestamp)
 
         # Retrieve cell numbers for the dependent symbols
         # Add dynamic and static dependencies
-        dep_cell_nums = cell_num_to_dynamic_deps[cell_num] | cell_num_to_static_deps[cell_num]
-        logger.info('dynamic cell deps for %d: %s', cell_num,
-                    cell_num_to_dynamic_deps[cell_num])
-        logger.info('static cell deps for %d: %s', cell_num,
-                    cell_num_to_static_deps[cell_num])
+        dep_timestamps = timestamp_to_dynamic_ts_deps[timestamp]
+        logger.info('dynamic ts deps for %s: %s', timestamp, dep_timestamps)
+        static_ts_deps = timestamp_to_static_ts_deps[timestamp]
+        dep_timestamps |= static_ts_deps
+        logger.info('static ts deps for %d: %s', timestamp, static_ts_deps)
 
         # For each dependent cell, recursively get their dependencies
-        for num in dep_cell_nums - dependencies:
-            self._get_cell_dependencies(
-                num, dependencies, cell_num_to_dynamic_deps, cell_num_to_static_deps)
+        for ts in dep_timestamps - dependencies:
+            self._get_ts_dependencies(
+                ts, dependencies, timestamp_to_dynamic_ts_deps, timestamp_to_static_ts_deps
+            )
 
     async def safe_execute(self, cell: str, is_async: bool, run_cell_func):
-        if self._saved_debug_message is not None:
+        if self._saved_debug_message is not None:  # pragma: no cover
             logger.error(self._saved_debug_message)
             self._saved_debug_message = None
         ret = None
@@ -706,26 +753,31 @@ class NotebookSafety(singletons.NotebookSafety):
                 self._counter_by_cell_id[cell_id] = self._last_execution_counter
                 self._cell_id_by_counter[self._last_execution_counter] = cell_id
 
+            if cell_id is not None:
+                # add to run_cells even if it fails precheck, since an attempt
+                # to run indicates that the user is interested in this cell, so
+                # we should add this cell to the set of cells that get checked
+                # in various ways in the post-exec hook.
+                self._run_cells.add(cell_id)
+
             # Stage 1: Precheck.
             if self._safety_precheck_cell(cell, cell_id) and self.settings.mark_stale_symbol_usages_unsafe:
                 # set this back in case we need it (e.g. user overrides and reruns)
                 self._active_cell_id = cell_id
                 # FIXME: hack to increase cell number
                 #  ideally we shouldn't show a cell number at all if we fail precheck since nothing executed
-                ret = run_cell_func('None')
                 if is_async:
-                    return await ret
+                    ret = await run_cell_func('None')  # pragma: no cover
                 else:
-                    return ret
+                    ret = run_cell_func('None')
+                return ret
 
             # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
             try:
-                if cell_id is not None:
-                    self._run_cells.add(cell_id)
                 self.cell_content_by_counter[self._last_execution_counter] = cell
                 with self._tracing_context(cell_id):
                     if is_async:
-                        ret = await run_cell_func(cell)
+                        ret = await run_cell_func(cell)  # pragma: no cover
                     else:
                         ret = run_cell_func(cell)
                 # Stage 2.1: resync any defined symbols that could have gotten out-of-sync
@@ -733,12 +785,13 @@ class NotebookSafety(singletons.NotebookSafety):
 
                 self._resync_symbols([
                     # TODO: avoid bad performance by only iterating over symbols updated in this cell
-                    sym for sym in self.all_data_symbols() if sym.timestamp == self.cell_counter()
+                    sym for sym in self.all_data_symbols() if sym.timestamp.cell_num == self.cell_counter()
                 ])
                 self._gc()
             except Exception as e:
-                if self.is_develop:
-                    logger.warning('Exception: %s', e)
+                # logger.exception('exception occurred during execution')
+                if self.is_test:
+                    self.set_exception_raised_during_execution(e)
             finally:
                 if not self.settings.store_history:
                     self._cell_counter += 1
@@ -763,7 +816,6 @@ class NotebookSafety(singletons.NotebookSafety):
     @contextmanager
     def _tracing_context(self, cell_id: CellId):
         self.updated_symbols.clear()
-        self._recorded_cell_name_to_cell_num = False
 
         try:
             with TraceManager.instance().tracing_context():
@@ -843,7 +895,7 @@ class NotebookSafety(singletons.NotebookSafety):
     def line_magic_name(self):
         return self._line_magic.__name__
 
-    def all_data_symbols(self):
+    def all_data_symbols(self) -> Iterable[DataSymbol]:
         for alias_set in self.aliases.values():
             yield from alias_set
 
