@@ -17,21 +17,6 @@ logger = logging.getLogger(__name__)
 logger.setLevel(logging.WARNING)
 
 
-def _maybe_convert_ast_subscript(subscript: ast.AST) -> ast.AST:
-    if isinstance(subscript, ast.Index):
-        return subscript.value  # type: ignore
-    elif isinstance(subscript, ast.Slice):
-        lower = fast.NameConstant(None) if subscript.lower is None else subscript.lower
-        upper = fast.NameConstant(None) if subscript.upper is None else subscript.upper
-        return fast.Call(
-            func=fast.Name('slice', ast.Load()),
-            args=[lower, upper] + ([] if subscript.step is None else [subscript.step]),
-            keywords=[],
-        )
-    else:
-        return subscript
-
-
 class AstEavesdropper(ast.NodeTransformer):
     def __init__(self, orig_to_copy_mapping: Dict[int, ast.AST]):
         self._orig_to_copy_mapping = orig_to_copy_mapping
@@ -91,14 +76,49 @@ class AstEavesdropper(ast.NodeTransformer):
             attr_or_sub = fast.Str(attr_node.attr)
         return self.visit_Attribute_or_Subscript(node, attr_or_sub, call_context=call_context)
 
+    def _maybe_convert_ast_subscript(self, subscript: ast.AST) -> ast.AST:
+        if isinstance(subscript, ast.Index):
+            return self.visit(subscript.value)  # type: ignore
+        elif isinstance(subscript, ast.Slice):
+            lower = fast.NameConstant(None) if subscript.lower is None else self.visit(subscript.lower)
+            upper = fast.NameConstant(None) if subscript.upper is None else self.visit(subscript.upper)
+            return fast.Call(
+                func=fast.Name('slice', ast.Load()),
+                args=[lower, upper] + ([] if subscript.step is None else [self.visit(subscript.step)]),
+                keywords=[],
+            )
+        else:
+            return subscript
+
     def visit_Subscript(self, node: ast.Subscript, call_context=False):
-        with fast.location_of(node.slice if hasattr(node.slice, 'lineno') else node.value):
-            attr_or_sub = _maybe_convert_ast_subscript(node.slice)
-            if isinstance(attr_or_sub, (ast.Slice, ast.ExtSlice)):
-                elts = attr_or_sub.elts if isinstance(attr_or_sub, ast.Tuple) else attr_or_sub.dims  # type: ignore
-                elts = [_maybe_convert_ast_subscript(elt) for elt in elts]
-                attr_or_sub = fast.Tuple(elts, ast.Load())
-        return self.visit_Attribute_or_Subscript(node, cast(ast.expr, attr_or_sub), call_context=call_context)
+        with self.attrsub_context(None):
+            with fast.location_of(node.slice if hasattr(node.slice, 'lineno') else node.value):
+                slc = self._maybe_convert_ast_subscript(node.slice)
+                if isinstance(slc, (ast.ExtSlice, ast.Tuple)):
+                    elts = slc.elts if isinstance(slc, ast.Tuple) else slc.dims  # type: ignore
+                    elts = [self._maybe_convert_ast_subscript(elt) for elt in elts]  # type: ignore
+                    slc = fast.Tuple(elts, ast.Load())
+                slc = fast.Call(
+                    func=self._emitter_ast(),
+                    args=[
+                        TraceEvent.subscript_slice.to_ast(),
+                        self._get_copy_id_ast(node.slice),
+                    ],
+                    keywords=fast.kwargs(ret=cast(ast.expr, slc)),
+                )
+                replacement_slice = fast.Call(
+                    func=self._emitter_ast(),
+                    args=[
+                        TraceEvent._load_saved_slice.to_ast(),
+                        self._get_copy_id_ast(node.slice),
+                    ],
+                    keywords=[],
+                )
+                if sys.version_info >= (3, 9):
+                    node.slice = replacement_slice
+                else:
+                    node.slice = fast.Index(replacement_slice)
+        return self.visit_Attribute_or_Subscript(node, slc, call_context=call_context)
 
     def _maybe_wrap_symbol_in_before_after_tracing(
         self, node, call_context=False, orig_node_id=None, begin_kwargs=None, end_kwargs=None
@@ -149,10 +169,9 @@ class AstEavesdropper(ast.NodeTransformer):
                         slice=orig_node.slice,
                     )
                 else:
-                    logger.error(
-                        'Symbol tracing stores unsupported for node %s with type %s', orig_node, type(orig_node)
+                    raise TypeError(
+                        'Symbol tracing stores unsupported for node %s with type %s' %(orig_node, type(orig_node))
                     )
-                    assert False
                 node.ctx = ast.Store()
         # end location_of(node)
         return node
