@@ -7,9 +7,10 @@ from nbsafety.analysis.symbol_edges import get_symbol_edges
 from nbsafety.analysis.utils import stmt_contains_lval
 from nbsafety.data_model.data_symbol import DataSymbol
 from nbsafety.data_model.scope import NamespaceScope
+from nbsafety.data_model.timestamp import Timestamp
 from nbsafety.singletons import nbs, tracer
 from nbsafety.tracing.mutation_event import ArgMutate, ListAppend, ListExtend, ListInsert
-from nbsafety.tracing.symbol_resolver import resolve_rval_symbols, update_usage_info
+from nbsafety.tracing.symbol_resolver import resolve_rval_symbols
 from nbsafety.tracing.utils import match_container_obj_or_namespace_with_literal_nodes
 
 if TYPE_CHECKING:
@@ -48,7 +49,9 @@ class TraceStatement:
         if isinstance(self.stmt_node, ast.ClassDef):
             # classes need a new scope before the ClassDef has finished executing,
             # so we make it immediately
-            return old_scope.make_child_scope(self.stmt_node.name, obj=-1)
+            pending_ns = old_scope.make_child_scope(self.stmt_node.name, obj=NamespaceScope.PENDING_CLASS_PLACEHOLDER)
+            tracer().pending_class_namespaces.append(pending_ns)
+            return pending_ns
 
         if isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
             func_name = self.stmt_node.name
@@ -81,7 +84,10 @@ class TraceStatement:
         except KeyError as e:
             # e.g., slices aren't implemented yet
             # use suppressed log level to avoid noise to user
-            logger.info("Exception: %s", e)
+            if nbs().is_develop:
+                logger.warning('keyerror for %s', ast.dump(target) if isinstance(target, ast.AST) else target)
+            # if nbs().is_test:
+            #     raise ke
             return
         upserted = scope.upsert_data_symbol_for_name(
             name, obj, deps, self.stmt_node, is_subscript=is_subscript,
@@ -181,7 +187,7 @@ class TraceStatement:
         assert isinstance(self.stmt_node, ast.Delete)
         for target in self.stmt_node.targets:
             try:
-                scope, name, _, is_subscript = tracer().resolve_del_data_for_target(target)
+                scope, obj, name, is_subscript = tracer().resolve_del_data_for_target(target)
                 scope.delete_data_symbol_for_name(name, is_subscript=is_subscript)
             except KeyError as e:
                 # this will happen if, e.g., a __delitem__ triggered a call
@@ -223,11 +229,17 @@ class TraceStatement:
                     class_scope=self.class_scope,
                     propagate=not isinstance(self.stmt_node, ast.For)
                 )
-            except KeyError:
-                logger.warning('keyerror for %s', ast.dump(target) if isinstance(target, ast.AST) else target)
+            except KeyError as ke:
+                # e.g., slices aren't implemented yet
+                # put logging behind flag to avoid noise to user
+                if nbs().is_develop:
+                    logger.warning('keyerror for %s', ast.dump(target) if isinstance(target, ast.AST) else target)
+                # if nbs().is_test:
+                #     raise ke
             except Exception as e:
                 logger.warning('exception while handling store: %s', e)
-                pass
+                if nbs().is_test:
+                    raise e
 
     def _handle_list_mutation(
         self,
@@ -264,13 +276,8 @@ class TraceStatement:
                     propagate=False,
                 )
         elif isinstance(mutation_event, ListInsert):
-            for idx in range(len(mutated_obj) - 1, mutation_event.insert_pos, -1):
-                subsym = namespace_scope._subscript_data_symbol_by_name.pop(idx - 1, None)
-                if subsym is None:
-                    continue
-                subsym.name = idx
-                namespace_scope._subscript_data_symbol_by_name[idx] = subsym
-                subsym.refresh()
+            assert mutated_obj is namespace_scope.obj
+            namespace_scope.shuffle_symbols_upward_from(mutation_event.insert_pos)
             namespace_scope.upsert_data_symbol_for_name(
                 mutation_event.insert_pos,
                 mutated_obj[mutation_event.insert_pos],
@@ -278,7 +285,7 @@ class TraceStatement:
                 self.stmt_node,
                 overwrite=False,
                 is_subscript=True,
-                propagate=False,
+                propagate=True,
             )
 
     def handle_dependencies(self):
@@ -287,10 +294,10 @@ class TraceStatement:
         for mutated_obj_id, mutation_event, mutation_arg_dsyms, mutation_arg_objs in tracer().mutations:
             propagate_to_namespace_descendents = not isinstance(mutation_event, (ListAppend, ListExtend, ListInsert))
             logger.info("mutation %s %s %s %s", mutated_obj_id, mutation_event, mutation_arg_dsyms, mutation_arg_objs)
-            update_usage_info(mutation_arg_dsyms)
+            Timestamp.update_usage_info(mutation_arg_dsyms)
             if isinstance(mutation_event, ArgMutate):
                 for mutated_sym in mutation_arg_dsyms:
-                    if mutated_sym is None:
+                    if mutated_sym is None or mutated_sym.is_anonymous:
                         continue
                     # TODO: happens when module mutates args
                     #  should we add module as a dep in this case?
@@ -305,7 +312,7 @@ class TraceStatement:
                 if isinstance(mutation_event, (ListAppend, ListInsert)):
                     mutation_arg_dsyms, mutation_upsert_deps = mutation_upsert_deps, mutation_arg_dsyms
                 self._handle_list_mutation(mutation_event, mutated_obj_id, mutation_upsert_deps)
-            update_usage_info(nbs().aliases[mutated_obj_id])
+            Timestamp.update_usage_info(nbs().aliases[mutated_obj_id])
             for mutated_sym in nbs().aliases[mutated_obj_id]:
                 mutated_sym.update_deps(
                     mutation_arg_dsyms,
