@@ -71,6 +71,28 @@ logging.info(f'The best model was {best_model} with an accuracy of {best_acc}.')
     assert_detected('`models` depends on stale `eval_model_1`')
 
 
+# TODO: to get this working properly, post-call argument
+#  symbols need to have dependencies on pre-call argument
+#  symbols. The place to do this is in `DataSymbol.create_symbols_for_call_args(...)`
+def test_passed_sym_captured_as_dep_for_mutated_obj():
+    run_cell("""
+class Foo:
+    def __init__(self, x):
+        self.x = x
+        
+def mutate(foo, x):
+    foo.x = x
+""")
+    run_cell('foo = Foo(5)')
+    run_cell('y = 7')
+    run_cell('mutate(foo, y)')
+    run_cell('logging.info(foo.x)')
+    assert_not_detected()
+    run_cell('y = 42')
+    run_cell('logging.info(foo.x)')
+    assert_detected('`foo.x` depends on old value of `y`')
+
+
 def test_subscript_dependency():
     run_cell('lst = [0, 1, 2]')
     run_cell('x = 5')
@@ -154,6 +176,55 @@ def test_lambda_params_not_live():
     assert_not_detected('`y` is not live in the lambda expr')
 
 
+def test_lambda_arg():
+    run_cell('x = 5')
+    run_cell('y = 7')
+    run_cell('lam = lambda x: x * 2')
+    run_cell('z = lam(y + 3)')
+    run_cell('x = 42')
+    run_cell('logging.info(z)')
+    assert_not_detected('`z` independent of updated `x`')
+    run_cell('y = 43')
+    run_cell('logging.info(z)')
+    assert_detected('`z` depends on old value of `y`')
+
+
+def test_lambda_scope():
+    run_cell("""
+def foo(x):
+    return lambda: x + 5
+""")
+    run_cell('x = 5')
+    run_cell('y = 7')
+    run_cell('lam = foo(y)')
+    run_cell('z = lam()')
+    run_cell('x = 42')
+    run_cell('logging.info(z)')
+    assert_not_detected('`z` independent of updated `x`')
+    run_cell('y = 43')
+    run_cell('logging.info(z)')
+    assert_detected('`z` depends on old value of `y`')
+
+
+def test_lambda_with_kwarg_scope():
+    run_cell('w = 99')
+    run_cell("""
+def foo(x):
+    lam = lambda t=w: t + x + 5
+    return lam
+""")
+    run_cell('x = 5')
+    run_cell('y = 7')
+    run_cell('lam = foo(y)')
+    run_cell('z = lam()')
+    run_cell('x = 42')
+    run_cell('logging.info(z)')
+    assert_not_detected('`z` independent of updated `x`')
+    run_cell('w = 43')
+    run_cell('logging.info(z)')
+    assert_detected('`z` depends on old value of `y`')
+
+
 def test_fundef_params_not_live():
     run_cell('x = 0')
     run_cell('y = x + 6')
@@ -560,9 +631,11 @@ def func(x, y=a):
     logging.info(b[0])
     e = [c[0] + d[0]]
     f = [x[0] + y[0]]
+    %safety show_deps f[0]
     return f
 # z = func(c)""")
     run_cell('z = func(c)')
+    run_cell('%safety show_deps z[0]')
     run_cell('logging.info(z[0])')
     assert_not_detected()
     run_cell('logging.info(z)')
@@ -1167,6 +1240,17 @@ def test_list_mutation():
     assert_detected('lst depends on stale x')
 
 
+def test_list_mutation_extend():
+    run_cell('lst = [0, 1]')
+    run_cell('x = lst[0] + 7')
+    run_cell('lst.extend([2, 3, 4, x])')
+    run_cell('logging.info(x)')
+    assert_not_detected()
+    run_cell('x = 77')
+    run_cell('logging.info(lst)')
+    assert_detected('`lst` depends on stale `x`')
+
+
 def test_list_mutation_2():
     run_cell('lst = list(range(5))')
     run_cell('x = lst + [42, 43]')
@@ -1187,6 +1271,17 @@ for word in s.split('X'):
     assert_detected('`lst` depends on stale `s`')
 
 
+@skipif_known_failing
+def test_list_mutation_extend_from_attr():
+    # for this one, we somehow have to avoid disabling tracing which is hard
+    run_cell('s = "hello X world X how X are X you X today?"')
+    run_cell('lst = []')
+    run_cell('lst.extend(word.strip() for word in s.split("X"))')
+    run_cell('s = "foobar"')
+    run_cell('logging.info(lst)')
+    assert_detected('`lst` depends on stale `s`')
+
+
 def test_lazy_class_scope_resolution():
     run_cell("""
 class Foo:
@@ -1447,6 +1542,27 @@ for i in range(5):
     assert_false_negative('`s` does depend on second entry of `lst` but tracing every iteration of loop is slow')
 
 
+def test_for_loop_tuple_unpack():
+    run_cell('x = (1, 2)')
+    run_cell('y = (3, 4)')
+    run_cell('total_i = 0')
+    run_cell('total_j = 0')
+    run_cell("""
+for i, j in [x, y]:
+    total_i += i
+    total_j += j
+""")
+    run_cell('logging.info(total_i)')
+    assert_not_detected()
+    run_cell('logging.info(total_j)')
+    assert_not_detected()
+    run_cell('x = (42, 43)')
+    run_cell('logging.info(total_i)')
+    assert_detected('`total_i` depends on old version of `x`')
+    run_cell('logging.info(total_j)')
+    assert_detected('`total_j` also depends on old version of `x`')
+
+
 def test_same_cell_redefine():
     run_cell('a = 0')
     run_cell("""
@@ -1606,7 +1722,7 @@ def test_no_class_false_positives():
     run_cell("""
 try:
     class Foo:
-        print(y)
+        logging.info(y)
 except:
     pass
 """)
@@ -2066,6 +2182,20 @@ def f(x, y):
     assert_detected('`w` depends on stale `foo.baz`')
 
 
+def test_list_extend():
+    run_cell('lst = [0, 1, 2]')
+    run_cell('x = lst[1] + 1')
+    run_cell('lst.extend([3, 4, 5, 6, 7, 8, 9])')
+    run_cell('logging.info(x)')
+    assert_not_detected()
+    run_cell('logging.info(lst[1])')
+    assert_not_detected()
+    run_cell('y = lst[8] + 1')
+    run_cell('lst[8] += 42')
+    run_cell('logging.info(y)')
+    assert_detected('`y` depends on stale `lst[8]`')
+
+
 @skipif_known_failing
 def test_list_sum_simple():
     run_cell("w, x, y, z = 0, 1, 2, 3")
@@ -2081,6 +2211,17 @@ def test_list_sum_simple():
     assert_detected()
 
 
+# TODO: where was I going with this?
+# def test_getitem_call():
+#     run_cell("""
+# class Foo:
+#     def __getitem__(self, x):
+#         return 42
+# """)
+#     run_cell('foo = Foo()')
+#     run_cell('x = foo[0]')
+
+
 if sys.version_info >= (3, 8):
     def test_walrus_simple():
         run_cell("""
