@@ -1,4 +1,5 @@
 # -*- coding: future_annotations -*-
+import difflib
 import logging
 import sys
 from typing import TYPE_CHECKING
@@ -8,17 +9,17 @@ from nbsafety.tracing.trace_manager import TraceManager
 from .utils import make_safety_fixture
 
 if TYPE_CHECKING:
-    from typing import Union
+    from typing import List, Union
     from types import FrameType
 
 logging.basicConfig(level=logging.INFO)
 
 
-RECORDED_EVENTS = []
+_RECORDED_EVENTS = []
 
 
 def patched_emit_event_fixture():
-    RECORDED_EVENTS.clear()
+    _RECORDED_EVENTS.clear()
     original_emit_event = TraceManager._emit_event
 
     def _patched_emit_event(self, evt: Union[TraceEvent, str], *args, **kwargs):
@@ -30,7 +31,7 @@ def patched_emit_event_fixture():
                 (event == TraceEvent.call and self.call_depth == 0) or
                 (event == TraceEvent.return_ and self.call_depth == 1)
             ):
-                RECORDED_EVENTS.append(event)
+                _RECORDED_EVENTS.append(event)
         return original_emit_event(self, evt, *args, **kwargs)
     TraceManager._emit_event = _patched_emit_event
     yield
@@ -44,6 +45,14 @@ _safety_fixture, run_cell_ = make_safety_fixture(
 )
 
 
+_DIFFER = difflib.Differ()
+
+
+def throw_and_print_diff_if_recorded_not_equal_to(actual: List[TraceEvent]) -> None:
+    assert _RECORDED_EVENTS == actual, '\n'.join(_DIFFER.compare(_RECORDED_EVENTS, actual))
+    _RECORDED_EVENTS.clear()
+
+
 def run_cell(cell, **kwargs):
     # print()
     # print('*******************************************')
@@ -54,9 +63,9 @@ def run_cell(cell, **kwargs):
 
 
 def test_recorded_events_simple():
-    assert RECORDED_EVENTS == []
+    assert _RECORDED_EVENTS == []
     run_cell('logging.info("foo")')
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.before_complex_symbol,
@@ -66,14 +75,15 @@ def test_recorded_events_simple():
         TraceEvent.after_call,
         TraceEvent.after_complex_symbol,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
+        TraceEvent.after_module_stmt,
+    ])
 
 
 def test_recorded_events_two_stmts():
-    assert RECORDED_EVENTS == []
+    assert _RECORDED_EVENTS == []
     run_cell('x = [1, 2, 3]')
     run_cell('logging.info(x)')
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.before_assign_rhs,
@@ -82,6 +92,7 @@ def test_recorded_events_two_stmts():
         TraceEvent.after_list_literal,
         TraceEvent.after_assign_rhs,
         TraceEvent.after_stmt,
+        TraceEvent.after_module_stmt,
 
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
@@ -92,13 +103,14 @@ def test_recorded_events_two_stmts():
         TraceEvent.after_call,
         TraceEvent.after_complex_symbol,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
+        TraceEvent.after_module_stmt,
+    ])
 
 
 def test_nested_chains_no_call():
-    assert RECORDED_EVENTS == []
+    assert _RECORDED_EVENTS == []
     run_cell('logging.info("foo is %s", logging.info("foo"))')
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.before_complex_symbol,
@@ -118,13 +130,14 @@ def test_nested_chains_no_call():
         TraceEvent.after_call,
         TraceEvent.after_complex_symbol,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
+        TraceEvent.after_module_stmt,
+    ])
 
 
 def test_list_nested_in_dict():
-    assert RECORDED_EVENTS == []
+    assert _RECORDED_EVENTS == []
     run_cell('x = {1: [2, 3, 4]}')
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.before_assign_rhs,
@@ -139,23 +152,24 @@ def test_list_nested_in_dict():
         TraceEvent.after_dict_literal,
         TraceEvent.after_assign_rhs,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
+        TraceEvent.after_module_stmt,
+    ])
 
 
 def test_function_call():
-    assert RECORDED_EVENTS == []
+    assert _RECORDED_EVENTS == []
     run_cell("""
 def foo(x):
     return [x]
 """)
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
-    RECORDED_EVENTS.clear()
+        TraceEvent.after_module_stmt,
+    ])
     run_cell('foo([42])')
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.before_complex_symbol,
@@ -175,13 +189,14 @@ def foo(x):
         TraceEvent.after_call,
         TraceEvent.after_complex_symbol,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
+        TraceEvent.after_module_stmt,
+    ])
 
 
 def test_lambda_in_tuple():
-    assert RECORDED_EVENTS == []
+    assert _RECORDED_EVENTS == []
     run_cell('x = (lambda: 42,)')
-    assert RECORDED_EVENTS == [
+    throw_and_print_diff_if_recorded_not_equal_to([
         TraceEvent.init_cell,
         TraceEvent.before_stmt,
         TraceEvent.before_assign_rhs,
@@ -192,4 +207,93 @@ def test_lambda_in_tuple():
         TraceEvent.after_tuple_literal,
         TraceEvent.after_assign_rhs,
         TraceEvent.after_stmt,
-    ], 'unexpected events; got %s' % RECORDED_EVENTS
+        TraceEvent.after_module_stmt,
+    ])
+
+
+def test_fancy_slices():
+    assert _RECORDED_EVENTS == []
+    run_cell("""
+import numpy as np
+class Foo:
+    def __init__(self, x):
+        self.x = x
+foo = Foo(1)
+arr = np.zeros((3, 3, 3))
+""")
+    throw_and_print_diff_if_recorded_not_equal_to([
+        TraceEvent.init_cell,
+        TraceEvent.before_stmt,
+        TraceEvent.after_stmt,
+        TraceEvent.after_module_stmt,
+
+        TraceEvent.before_stmt,
+        TraceEvent.call,
+        TraceEvent.before_stmt,
+        TraceEvent.after_stmt,
+        TraceEvent.return_,
+        TraceEvent.after_stmt,
+        TraceEvent.after_module_stmt,
+
+        TraceEvent.before_stmt,
+        TraceEvent.before_assign_rhs,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.before_call,
+        TraceEvent.argument,
+        TraceEvent.call,
+        TraceEvent.before_stmt,
+        TraceEvent.before_assign_rhs,
+        TraceEvent.after_assign_rhs,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.attribute,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.after_stmt,
+        TraceEvent.return_,
+        TraceEvent.after_call,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.after_assign_rhs,
+        TraceEvent.after_stmt,
+        TraceEvent.after_module_stmt,
+
+        TraceEvent.before_stmt,
+        TraceEvent.before_assign_rhs,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.attribute,
+        TraceEvent.before_call,
+        TraceEvent.before_tuple_literal,
+        TraceEvent.tuple_elt,
+        TraceEvent.tuple_elt,
+        TraceEvent.tuple_elt,
+        TraceEvent.after_tuple_literal,
+        TraceEvent.argument,
+        TraceEvent.after_call,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.after_assign_rhs,
+        TraceEvent.after_stmt,
+        TraceEvent.after_module_stmt,
+    ])
+
+    run_cell('logging.info(arr[foo.x:foo.x+1,...])')
+    throw_and_print_diff_if_recorded_not_equal_to([
+        TraceEvent.init_cell,
+        TraceEvent.before_stmt,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.attribute,
+        TraceEvent.before_call,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.attribute,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.before_complex_symbol,
+        TraceEvent.attribute,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.subscript_slice,
+        TraceEvent.subscript,
+        TraceEvent._load_saved_slice,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.argument,
+        TraceEvent.after_call,
+        TraceEvent.after_complex_symbol,
+        TraceEvent.after_stmt,
+        TraceEvent.after_module_stmt,
+    ])
